@@ -1,5 +1,5 @@
 /* hdparm.c - Command line interface to get/set hard disk parameters */
-/*          - by Mark S. Lord (c) 1994-1997 -- freely distributable */
+/*          - by Mark S. Lord (c) 1994-1998 -- freely distributable */
 
 #include <unistd.h>
 #include <string.h>
@@ -29,7 +29,7 @@
 #endif
 #endif
 
-#define TIMING_MB		32
+#define TIMING_MB		64
 #define TIMING_BUF_MB		1
 #define TIMING_BUF_BYTES	(TIMING_BUF_MB * 1024 * 1024)
 #define TIMING_BUF_COUNT	(TIMING_MB / TIMING_BUF_MB)
@@ -39,7 +39,7 @@ char *progname;
 static struct hd_driveid id;
 static int verbose = 0, get_identity = 0, get_geom = 0, noisy = 1, quiet = 0;
 static int flagcount = 0, do_flush = 0, is_scsi_hd = 0;;
-static int do_ctimings, do_timings = 0;
+static int do_ctimings, do_timings = 0, old_identity = 0;
 
 static unsigned long set_readahead= 0, get_readahead= 0, readahead= 0;     
 static unsigned long set_readonly = 0, get_readonly = 0, readonly = 0;     
@@ -66,6 +66,7 @@ static unsigned long set_standby  = 0, get_standby  = 0, standby  = 0;
 static unsigned long set_xfermode = 0, get_xfermode = 0, xfermode = 0;
 static unsigned long set_lookahead= 0, get_lookahead= 0, lookahead= 0;
 static unsigned long set_prefetch = 0, get_prefetch = 0, prefetch = 0;
+static unsigned long set_defects  = 0, defects = 0;
 static unsigned long set_wcache   = 0, get_wcache   = 0, wcache   = 0;
 static unsigned long set_seagate  = 0, get_seagate  = 0;
 static unsigned long set_standbynow = 0, get_standbynow = 0;
@@ -96,7 +97,7 @@ static void dump_identity (struct hd_driveid *id)
 {
 	int i;
 
-	printf("\n Model=%.40s, FwRev=%.8s, SerialNo=%.8s", 
+	printf("\n Model=%.40s, FwRev=%.8s, SerialNo=%.20s", 
 		id->model, id->fw_rev, id->serial_no);
 	printf("\n Config={");
 	for (i=0; i<=15; i++) {
@@ -129,8 +130,27 @@ static void dump_identity (struct hd_driveid *id)
 	printf(" CurCHS=%d/%d/%d, CurSects=%d",
 		id->cur_cyls, id->cur_heads, id->cur_sectors, *(int *)&id->cur_capacity0);
 	printf(", LBA=%s", YN(id->capability&2));
-	if (id->capability&2)
-		printf(", LBAsects=%d", id->lba_capacity);
+	if (id->capability&2) {
+		if ((id->cur_sectors == 63) && (id->cur_cyls >= 1024)) {
+			unsigned short new_cyls  = id->cur_cyls;
+			unsigned short new_heads = id->cur_heads;
+
+			if (id->cur_heads == 15) new_heads++;
+
+			do {
+				new_cyls /= 2;
+				new_heads *= 2;
+			} while (new_cyls >=1024);
+
+			if (id->cur_heads == 15)
+				new_heads--;
+
+			printf("\n LBA CHS=%d/%d/%d Remapping",
+				((*(int *)&id->cur_capacity0)/(new_heads * id->cur_sectors)), new_heads, id->cur_sectors);
+		}
+		printf(", LBA=%s", YN(id->capability&2));
+ 		printf(", LBAsects=%d", id->lba_capacity);
+	}
 	if (id->capability&1) {
 		if ((id->field_valid&2) || (id->dma_1word | id->dma_mword))
 			printf("\n ");
@@ -169,9 +189,26 @@ static void dump_identity (struct hd_driveid *id)
 		if (id->field_valid&2) {
 			printf(", tPIO={min:%d,w/IORDY:%d}, PIO modes: ",
 			 id->eide_pio, id->eide_pio_iordy);
-			if (id->eide_pio_modes & 1) printf("mode3 ");
-			if (id->eide_pio_modes & 2) printf("mode4 ");
-			if (id->eide_pio_modes &~3) printf("mode? ");
+			if (id->eide_pio_modes & 1)	printf("mode3 ");
+			if (id->eide_pio_modes & 2)	printf("mode4 ");
+			if (id->eide_pio_modes &~3)	printf("mode? ");
+		}
+		if (id->field_valid&4) {
+			printf("\n UDMA modes: ");
+#ifndef HDIO_OBSOLETE_IDENTITY
+			printf(" (not available: rebuild hdparm with a newer kernel");
+#else
+			if (old_identity) {
+				printf(" (not available from this kernel)");
+			} else {
+				if (id->dma_ultra & 0x100)	printf("*");
+				if (id->dma_ultra & 0x001)	printf("mode0 ");
+				if (id->dma_ultra & 0x200)	printf("*");
+				if (id->dma_ultra & 0x002)	printf("mode1 ");
+				if (id->dma_ultra & 0x400)	printf("*");
+				if (id->dma_ultra & 0x004)	printf("mode2");
+			}
+#endif
 		}
 	}
 	printf("\n\n");
@@ -447,12 +484,18 @@ static void interpret_xfermode (unsigned int xfermode)
 		case 38:
 		case 39:	printf("multiword DMA mode%d", xfermode-32);
 				break;
+		case 64:
+		case 65:
+		case 66:
+		case 67:
+		case 68:
+		case 69:
+		case 70:
+		case 71:	printf("UltraDMA mode%d", xfermode-64);
+				break;
 		default:
-			if ((xfermode & 0x78) == 64)
-				printf("reserved mode%d", xfermode-64);
-			else
 				printf("unknown, probably not valid");
-			break;
+				break;
 	}
 	printf(")\n");
 }
@@ -471,11 +514,22 @@ void process_dev (char *devname)
 		exit(errno);
 	}
 
+#ifdef SCSI_DISK0_MAJOR
+	if ((major(stat_buf.st_rdev) == SCSI_DISK0_MAJOR) 
+	 || (major(stat_buf.st_rdev) == SCSI_DISK1_MAJOR) 
+	 || (major(stat_buf.st_rdev) == SCSI_DISK2_MAJOR) 
+	 || (major(stat_buf.st_rdev) == SCSI_DISK3_MAJOR) 
+	 || (major(stat_buf.st_rdev) == SCSI_DISK4_MAJOR) 
+	 || (major(stat_buf.st_rdev) == SCSI_DISK5_MAJOR) 
+	 || (major(stat_buf.st_rdev) == SCSI_DISK6_MAJOR) 
+	 || (major(stat_buf.st_rdev) == SCSI_DISK7_MAJOR) 
+#else
 	if ((major(stat_buf.st_rdev) == SCSI_DISK_MAJOR) 
-#ifdef MD_MAJOR
-		|| (major(stat_buf.st_rdev) == MD_MAJOR)
 #endif
-							) 
+#ifdef MD_MAJOR
+	 || (major(stat_buf.st_rdev) == MD_MAJOR)
+#endif
+	) 
 		is_scsi_hd = 1;
 	else
 		if (major(stat_buf.st_rdev) != IDE0_MAJOR
@@ -600,6 +654,15 @@ void process_dev (char *devname)
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
 			perror(" HDIO_DRIVE_CMD failed");
 	}
+	if (set_defects) {
+		unsigned char args[4] = {WIN_SETFEATURES,0,0x04,0};
+		no_scsi();
+		args[2] = defects ? 0x04 : 0x84;
+		if (get_prefetch)
+			printf(" setting drive defect-mgmt to %ld\n", defects);
+		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
+			perror(" HDIO_DRIVE_CMD failed");
+	}
 	if (set_prefetch) {
 		unsigned char args[4] = {WIN_SETFEATURES,0,0xab,0};
 		no_scsi();
@@ -643,21 +706,27 @@ void process_dev (char *devname)
 			perror(" HDIO_DRIVE_CMD failed");
 	}
 	if (set_standbynow) {
-#define WIN_STANDBYNOW 0x94	/* or 0xE0 */
-		unsigned char args[4] = {WIN_STANDBYNOW,0,0,0};
+#define WIN_STANDBYNOW1 0xE0
+#define WIN_STANDBYNOW2 0x94
+		unsigned char args1[4] = {WIN_STANDBYNOW1,0,0,0};
+		unsigned char args2[4] = {WIN_STANDBYNOW2,0,0,0};
 		no_scsi();
 		if (get_standbynow)
 			printf(" issuing standby command\n");
-		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
+		if (ioctl(fd, HDIO_DRIVE_CMD, &args1)
+		 && ioctl(fd, HDIO_DRIVE_CMD, &args2))
 			perror(" HDIO_DRIVE_CMD failed");
 	}
 	if (set_sleepnow) {
-#define WIN_SLEEPNOW 0x99	/* or 0xE6 */
-		unsigned char args[4] = {WIN_SLEEPNOW,0,0,0};
+#define WIN_SLEEPNOW1 0xE6
+#define WIN_SLEEPNOW2 0x99
+		unsigned char args1[4] = {WIN_SLEEPNOW1,0,0,0};
+		unsigned char args2[4] = {WIN_SLEEPNOW2,0,0,0};
 		no_scsi();
 		if (get_sleepnow)
 			printf(" issuing sleep command\n");
-		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
+		if (ioctl(fd, HDIO_DRIVE_CMD, &args1)
+		 && ioctl(fd, HDIO_DRIVE_CMD, &args2))
 			perror(" HDIO_DRIVE_CMD failed");
 	}
 	if (set_seagate) {
@@ -790,29 +859,32 @@ void process_dev (char *devname)
 	}
 #ifdef HDIO_DRIVE_CMD
 	if (get_powermode) {
-#define WIN_CHECKPOWERMODE	0x98	/* or 0xE5 */
-		unsigned char args[4] = {WIN_CHECKPOWERMODE,0,0,0};
+#define WIN_CHECKPOWERMODE1 0xE5
+#define WIN_CHECKPOWERMODE2 0x98
+		unsigned char args[4] = {WIN_CHECKPOWERMODE1,0,0,0};
 		const char *state;
 		no_scsi();
-		if (ioctl(fd, HDIO_DRIVE_CMD, &args)) {
-			if (errno != EIO || args[0] != 0 || args[1] != 0) {
-				/* perror(" HDIO_DRIVE_CMD failed"); */
+		if (ioctl(fd, HDIO_DRIVE_CMD, &args)
+		 && (args[0] = WIN_CHECKPOWERMODE2) /* try again with 0x98 */
+		 && ioctl(fd, HDIO_DRIVE_CMD, &args)) {
+			if (errno != EIO || args[0] != 0 || args[1] != 0)
 				state = "unknown";
-			} else {
-				state = "sleeping";
-			}
-		} else {
-			if (args[2] == 255)
-				state = "active/idle";
 			else
-				state = "standby";
+				state = "sleeping";
+		} else {
+			state = (args[2] == 255) ? "active/idle" : "standby";
 		} 
 		printf(" drive state is:  %s\n", state);
 	}
 #endif
 	if (get_identity) {
 		no_scsi();
-		if (!(ioctl(fd, HDIO_GET_IDENTITY, &id))) {
+#ifdef HDIO_OBSOLETE_IDENTITY
+		if (!(old_identity = ioctl(fd, HDIO_GET_IDENTITY, &id)) || !ioctl(fd, HDIO_OBSOLETE_IDENTITY)) {
+#else
+		if (!ioctl(fd, HDIO_GET_IDENTITY, &id)) {
+			old_identity = 1;
+#endif
 			if (multcount != -1) {
 				id.multsect = multcount;
 				id.multsect_valid |= 1;
@@ -826,12 +898,18 @@ void process_dev (char *devname)
 	}
 #ifdef HDIO_SET_32BIT
 	if (get_IDentity) {
-		unsigned char args[4+512] = {WIN_IDENTIFY,0,0,1,};
+		unsigned char args[4+512] = {WIN_PIDENTIFY,0,0,1,};
 		no_scsi();
-		if (!(ioctl(fd, HDIO_DRIVE_CMD, &args)))
+		if (!(ioctl(fd, HDIO_DRIVE_CMD, &args))) {
+			old_identity = 0;
 			dump_identity((struct hd_driveid *)&args[4]);
-		else
-			perror(" HDIO_DRIVE_CMD failed");
+		} else {
+			args[0] = WIN_IDENTIFY;
+			if (!(ioctl(fd, HDIO_DRIVE_CMD, &args)))
+				dump_identity((struct hd_driveid *)&args[4]);
+			else
+				perror(" HDIO_DRIVE_CMD failed");
+		}
 	}
 #endif
 	if (do_ctimings)
@@ -845,7 +923,7 @@ void process_dev (char *devname)
 
 void usage_error (void)
 {
-	fprintf(stderr,"\n%s - get/set hard disk parameters - version 3.3\n\n", progname);
+	fprintf(stderr,"\n%s - get/set hard disk parameters - version 3.5\n\n", progname);
 	fprintf(stderr,"Usage:  %s  [options] [device] ..\n\n", progname);
 	fprintf(stderr,"Options:\n"
 	" -a    get/set fs readahead\n"
@@ -861,6 +939,7 @@ void usage_error (void)
 #ifdef HDIO_SET_DMA
 	" -d  * get/set using_dma flag\n"
 #endif
+	" -D  * enable/disable drive defect-mgmt\n"
 	" -f    flush buffer cache for device on exit\n"
 	" -g    display drive geometry\n"
 	" -h    display terse usage information\n"
@@ -896,7 +975,7 @@ void usage_error (void)
 #ifdef HDIO_DRIVE_CMD
 	" -W  * set drive write-caching flag (0/1) (DANGEROUS)\n"
 	" -X  * set IDE xfer mode (DANGEROUS)\n"
-	" -Y  * put IDE drive in standby mode\n"
+	" -y  * put IDE drive in standby mode\n"
 	" -Y  * put IDE drive to sleep\n"
 	" -Z  * disable Seagate auto-powersaving mode\n"
 #endif
@@ -1031,6 +1110,14 @@ int main(int argc, char **argv)
 							fprintf(stderr, "-S: missing value\n");
 						break;
 
+					case 'D':
+						set_defects = noisy;
+						noisy = 1;
+						GET_NUMBER(set_defects,defects);
+						if (!set_defects)
+							fprintf(stderr, "-D: missing value\n");
+						break;
+
 					case 'P':
 						get_prefetch = noisy;
 						noisy = 1;
